'use strict';

let thunkify = require('thunkify');
let co = require('co');
let parallel = require('co-parallel');
let mongo=require("mongodb");
let BemZk = require('bem-zk');
let zkAPI;

/* record each area  ArrearageList, it is "2 usera user b"*/
let ArrearageList = {};
let criticalData={};
let normalData={mongoDbIP:'127.0.0.1', /* mongoDb host IP */
                mongoDbPort:27017, /* mongoDb port */
                CountMDPort:12347,/* max buffer user number call API, if get it, must send data out. */
                maxParallelNumber:30,
                maxApiTypeNumber:10, /* API type number, max is 10 so far. */
                intervalTimer:60};/* update arrange list timer, this should not need be a configure data. */
let intervalId;
let dbOpen;
 
 /*
 *----------------------------------------------------------------------
 *
 *  updateDailyTable(jsonData)
 *
 *  this is a update of userCountDailyHistory
 *----------------------------------------------------------------------
 */
function *updateDailyTable(jsonData)
{
    let date  = new Date();

    let todayYear  = date.getFullYear();
    let todayMonth = date.getMonth();
    let todayDay   = date.getDate();
    let updateData = {"count":jsonData.count};
    let timeStamp=new Date(todayYear,todayMonth,todayDay,0,0,0,0).getTime();
    let findResult;
    let docs ;

    delete  jsonData["_id"];

    let userCountDailyHistoryCollection = dbOpen.collection('userCountDailyHistory');
    try
    {
        findResult = yield thunkify(userCountDailyHistoryCollection.find).call(userCountDailyHistoryCollection,
                                    {userId:jsonData.userId,date:{"$gt":timeStamp}});
        docs = yield thunkify(findResult.toArray).call(findResult);
    }
    catch(err)
    {
        console.log(err);
    }

    if(docs.length==0)
    {
        /* insert */
        yield thunkify(userCountDailyHistoryCollection.insert).call(userCountDailyHistoryCollection,jsonData);

    }
    else
    {

        for(let i=0;i<normalData.maxApiTypeNumber;i++)
        {
            if(jsonData[i.toString()]!=undefined)
            {
                updateData[i.toString()]=jsonData[i.toString()];
            }
        }
        let result = yield thunkify(userCountDailyHistoryCollection.update).call(userCountDailyHistoryCollection,
                                        {userId:jsonData.userId,appId:jsonData.appId,date:{"$gt":timeStamp}}, 
                                        {$inc:updateData});
    }

    return;
}

 /*
 *----------------------------------------------------------------------
 *
 *  updateDailyDbTable(jsonData)
 *
 *  check does this user's daily data already exist in.
 *      1)       exist: call updateDailyTable()
 *      1)don't exist: call instertDailyTable()
 *----------------------------------------------------------------------
 */
function  updateDailyDbTable(jsonData)
{
    co(function* () {
          let reqs = jsonData.map(updateDailyTable);
          let res = yield parallel(reqs, normalData.maxParallelNumber);
          });
    
    return;
}

 /*
 *----------------------------------------------------------------------
 *
 *  updateCount2Db(jasonData)
 *
 *    This routine will update count to db
 *----------------------------------------------------------------------
 */
function updateCount2Db(jsonData)
{
    let date  = new Date();
    let userId;
    let appId;
    let parseKeyName;

    for(let j=0;j<jsonData.length;j++)
    {
        parseKeyName = jsonData[j].keyName.match(/([^_]+)_.*?(.+)/);
        jsonData[j].userId=parseKeyName[1];
        jsonData[j].appId=parseKeyName[2];
        delete jsonData[j].keyName;
        /* add timestamp to this json. */
        jsonData[j].date=date.getTime();
        
    }

    dbOpen.collection("userCountHistory", function (err,collection) {
        if(err)
        {
            throw err;
        }
        else{
            /* insert this session all data to  userCountHistory.*/
            collection.insert(jsonData, function (err,docs) {
      		if(err) 
      		{
      			/* here has error, should exit acountMD to make sure countSD will not lost data. */
      			console.log(" insert error in updateCount2Db.");
      			//console.log(jasonData[0]);
      			throw err;
      		}
            });
        }
    });

    /* need update  Daily data. */
    updateDailyDbTable(jsonData);
    
    return;
}
/* 
 *----------------------------------------------------------------------
 *
 * criticalChange()
 * called when critical change happen in zookeeper.
 *    INPUT: jsonData    : bemcloud apicount udp server port number.
 *    INPUT: countServerPort  : countD serverport number.
 *----------------------------------------------------------------------
 */
function countMDcriticalChange(jsonData)
{
    //console.log("criticalChange is called in countAPI and data is.");
    //console.log(jsonData);

    return;
}

/* 
 *----------------------------------------------------------------------
 *
 *  countAPInormalDataChange(jsonData)
 *  called when normal change happen in zookeeper.
 *    INPUT: jsonData    : changed data.
 *----------------------------------------------------------------------
 */

function countMDnormalDataChange(jsonData)
{
    //console.log("countAPInormalDataChange is called in countAPI and data is.");
    //console.log(jsonData);


    if((jsonData.mongoDbIP!=undefined)&&(jsonData.mongoDbIP!=normalData.mongoDbIP))
    { 
        /* this happen, we need restart. */
        process.exit(0);
        return;
    }

    if((jsonData.mongoDbPort!=undefined)&&(jsonData.mongoDbPort!=normalData.mongoDbPort))
    {
         /* this happen, we need restart. */
        process.exit(0);
        return;
    }
    
    if((jsonData.CountMDPort!=undefined)&&(jsonData.CountMDPort!=normalData.CountMDPort))
    {
        /* this happen, restart is more easy. */
        process.exit(0);        
        return;
    }
    
    if((jsonData.maxParallelNumber!=undefined)&&(jsonData.maxParallelNumber!=normalData.maxParallelNumber))
    {
        /* just change it. */
        normalData.maxParallelNumber = jsonData.maxParallelNumber;
        return;
    }
      
    if((jsonData.maxApiTypeNumber!=undefined)&&(jsonData.maxApiTypeNumber!=normalData.maxApiTypeNumber))
    {
        /* maxApiTypeNumber is changed, just change it. */
        normalData.maxApiTypeNumber = jsonData.maxApiTypeNumber;
        return;
    }

    if((jsonData.intervalTimer!=undefined)&&(jsonData.intervalTimer!=normalData.intervalTimer))
    {
        /* timer is changed, reset timer. */
        updateArrearageList();
        clearInterval(intervalId);
        normalData.intervalTimer = jsonData.intervalTimer;
        intervalId = setInterval(updateArrearageList, normalData.intervalTimer*1000);
        return;
    }

   return;
}

 /*
 *----------------------------------------------------------------------
 *
 * initAccountServer()
 *
 *    This routineshoule is called when accountD is init,
 *  INPUT accountServerPort : accountD tcp server port.
 *----------------------------------------------------------------------
 */
function *initCountMDServer()
{
    let moduleName="countMD";
    let getData;
    
    zkAPI = new BemZk();
    zkAPI.initZookeeper(moduleName,countMDcriticalChange,countMDnormalDataChange);

    getData = yield zkAPI.getNormalZkData(moduleName,"mongoDbIP");
    if(getData!=undefined)
    {
        normalData.mongoDbIP = getData;
    }
    getData = yield zkAPI.getNormalZkData(moduleName,"mongoDbPort");
    if(getData!=undefined)
    {
        normalData.mongoDbPort = getData;
    }
    getData = yield zkAPI.getNormalZkData(moduleName,"CountMDPort");
    if(getData!=undefined)
    {
        normalData.CountMDPort = getData;
    }
    getData = yield zkAPI.getNormalZkData(moduleName,"maxParallelNumber");
    if(getData!=undefined)
    {
        normalData.maxParallelNumber = getData;
    }
    getData = yield zkAPI.getNormalZkData(moduleName,"maxApiTypeNumber");
    if(getData!=undefined)
    {
        normalData.maxApiTypeNumber = getData;
    }

    console.log(normalData);

    let server=new mongo.Server(normalData.mongoDbIP,parseInt(normalData.mongoDbPort),{auto_reconnect:true});
    let db=new mongo.Db("_BEMCLOUD",server,{safe:true});
    dbOpen = yield thunkify(db.open).call(db);
    /* update ArrearageList first. */
    updateArrearageList();
    /*init tcp server*/
    let tcpServer = require('net').createServer(function(socket) {
        console.log('new connection');
        socket.setEncoding('utf8');
        socket.on('data', function(receiveString) {
            let jsonData 
             console.log("connetc success in tcp.");
                 try{
                   jsonData =JSON.parse(receiveString);
                 } catch (err) {
                    console.log("Server got: wrong data"+receiveString);
                    socket.end();
                    return;
                 }
                 if(ArrearageList[jsonData[0].area]==undefined)
                 {
                    ArrearageList[jsonData[0].area]="0";
                    console.log('error happen from count, there is no area in this count');
                 }
                 console.log('reply tcp data is :' +ArrearageList[jsonData[0].area]);
                 console.log(typeof(jsonData[0].area));
                 socket.write(ArrearageList[jsonData[0].area]);
                 socket.end();
                 updateCount2Db(jsonData);
        });
        
        socket.on('end', function() {
                console.log('Client connection ended');
         });

     }).listen(normalData.CountMDPort);

    return;
}

 /*
 *----------------------------------------------------------------------
 *
 * updateArrearageListForArea()
 *
 *    This routineshoule is called by updateArrearageList in timer callback(1 minute)
 *  will get all this areaIndex Arrearage userName to a string ArrearageList[areaIndex] 
 *
 *  ArrearageList[areaIndex]  format is :  "arrageNumber apiBufferNumber ArrearageName1 ArrearageName2 ..."
 *  INPUT accountServerPort : area Index.
 *        apiBufferNumber   : this area apiBufferNumber.
 *----------------------------------------------------------------------
 */
function updateArrearageListForArea(areaIndex)
{
    dbOpen.collection("user", function (err,collection) {
      if(err) throw err;
      else{
          collection.find( {inService:0,area:areaIndex},{"userId":1,_id:0}).toArray(function(err,docs){
            if(err) throw  err;
            else{
                let len = docs.length;
                /* arrageNumber + apiBufferNumber +  ArrearageList */
                ArrearageList[areaIndex.toString()] = docs.length.toString() ;
                let i = 0;
                for (; i < len; ++i) {
                    ArrearageList[areaIndex.toString()] = ArrearageList[areaIndex.toString()]  + ' ' +docs[i].userId;
                }
                if (len == 0) {
                    ArrearageList[areaIndex.toString()] = "0";
                }
                console.log("update from db, ArrearageList is");
                console.log(ArrearageList);
            }
        });
      }
    });

   return;
}

 /*
 *----------------------------------------------------------------------
 *
 * updateArrearageList()
 *
 *    This routineshoule is called each timer callback(1 minute)
 *  will update all ArrearageList[] 
 *----------------------------------------------------------------------
 */
function updateArrearageList()
{
    dbOpen.collection("areaInfo", function (err,areaInfoCollection) {
        /* get APIBufferNumber and area id out. */
        areaInfoCollection.find( {},{}).toArray(function(err,docs){
            if(err) throw  err;
            else{
                
                let len = docs.length,
                i = 0;
                for (; i < len; ++i) 
                {
                    updateArrearageListForArea(docs[i].area);
                }
             }
        });
    });  

   return;
}

 /*
 *----------------------------------------------------------------------
 *
 * main()
 *
 *    This routineshoule is countMD, it is used for update all users counter in our main server.
 *  INPUT 1 parameter : tcp server port number.
 *----------------------------------------------------------------------
 */
function main() {

    co(function *(){
        yield initCountMDServer();
    }).then(
        intervalId=setInterval(updateArrearageList, normalData.intervalTimer*1000)
    );
    
    return;
}

main();


